#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
from random import choice
from time import time

import requests

from octopus import Octopus


def main(repetitions, concurrency):

    # alexa top sites
    urls = [
        'http://google.com',
        'http://facebook.com',
        'http://youtube.com',
        'http://yahoo.com',
        'http://wikipedia.org',
        'http://linkedin.com',
        'http://live.com',
        'http://twitter.com',
        'http://amazon.com',
        'http://blogspot.com',
        'http://wordpress.com',
        'http://bing.com',
        'http://ebay.com',
        'http://google.de',
        'http://tumblr.com',
    ]

    urls_to_retrieve = [choice(urls) for i in range(repetitions)]

    message = "Retrieving URLs sequentially with Requests..."
    print
    print("=" * len(message))
    print(message)
    print("=" * len(message))
    print

    start_time = time()

    for url_index, url in enumerate(urls_to_retrieve):
        print "%.2f%% - getting %s..." % (
            float(url_index) / float(repetitions) * 100,
            url
        )
        assert requests.get(url).status_code == 200

    requests_total_time = time() - start_time

    message = "Retrieving URLs concurrently with Octopus..."
    print
    print("=" * len(message))
    print(message)
    print("=" * len(message))
    print

    otto = Octopus(concurrency=concurrency)

    for url in urls_to_retrieve:
        otto.enqueue(url, handle_url_response)

    start_time = time()
    otto.start()
    otto.wait()

    otto_total_time = time() - start_time

    message = "Retrieving URLs concurrently with Octopus with caching enabled..."
    print
    print("=" * len(message))
    print(message)
    print("=" * len(message))
    print

    otto = Octopus(concurrency=concurrency, cache=True, auto_start=True)

    for url in urls_to_retrieve:
        otto.enqueue(url, handle_url_response)

    start_time = time()
    otto.wait()

    otto_cached_total_time = time() - start_time

    message = "RESULTS"
    print
    print("=" * len(message))
    print(message)
    print("=" * len(message))
    print

    print "[requests] Retrieving %d urls took %.2f seconds meaning %.2f urls/second" % (
        repetitions,
        requests_total_time,
        repetitions / requests_total_time
    )
    print

    print "[octopus] Retrieving %d urls took %.2f seconds meaning %.2f urls/second" % (
        repetitions,
        otto_total_time,
        repetitions / otto_total_time
    )
    print

    print "[octopus] Retrieving %d urls with local in-memory caching took %.2f seconds meaning %.2f urls/second" % (
        repetitions,
        otto_total_time,
        repetitions / otto_total_time
    )
    print

    print "Overall, octopus was more than %.2f times faster than sequential requests." % (
        int(requests_total_time / otto_total_time)
    )

    print


def handle_url_response(url, response):
    print "Got %s!" % url
    assert response.status_code == 200, "Expected status code for %s to be 200, got %s" % (url, response.status_code)


if __name__ == '__main__':
    main(int(sys.argv[1]), int(sys.argv[2]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys

from preggy import expect
from mock import Mock

from octopus import Octopus, TimeoutError
from tests import TestCase


class TestOctopus(TestCase):
    def setUp(self):
        self.response = None
        self.responses = {}

    def test_can_create_octopus(self):
        otto = Octopus(concurrency=20)
        expect(otto.concurrency).to_equal(20)
        expect(otto.auto_start).to_be_false()
        expect(otto.cache).to_be_false()

    def test_has_default_concurrency(self):
        otto = Octopus()
        expect(otto.concurrency).to_equal(10)

    def test_queue_is_empty(self):
        otto = Octopus()
        expect(otto.is_empty).to_be_true()

    def test_can_enqueue_url(self):
        otto = Octopus()

        otto.enqueue('http://www.google.com', None)

        expect(otto.queue_size).to_equal(1)

    def test_can_get_after_started(self):
        otto = Octopus(concurrency=1)

        def handle_url_response(url, response):
            self.response = response

        otto.enqueue('http://www.google.com', handle_url_response)
        otto.start()

        otto.wait(2)

        expect(self.response).not_to_be_null()
        expect(self.response.status_code).to_equal(200)

    def test_can_get_with_auto_start(self):
        otto = Octopus(concurrency=1, auto_start=True)

        def handle_url_response(url, response):
            self.response = response

        otto.enqueue('http://www.google.com', handle_url_response)

        otto.wait(2)

        expect(self.response).not_to_be_null()
        expect(self.response.status_code).to_equal(200)

    def test_can_wait(self):
        otto = Octopus(concurrency=1)

        def handle_url_response(url, response):
            self.response = response

        otto.enqueue('http://www.google.com', handle_url_response)
        otto.start()

        otto.wait(0)

        expect(self.response).not_to_be_null()
        expect(self.response.status_code).to_equal(200)

    def test_times_out_on_wait(self):
        otto = Octopus(concurrency=1)

        def handle_url_response(url, response):
            self.response = response

        otto.enqueue('http://www.google.com', handle_url_response)

        try:
            otto.wait(0.1)
        except TimeoutError:
            err = sys.exc_info()[1]
            expect(err).to_have_an_error_message_of("")
        else:
            assert False, "Should not have gotten this far"

    def test_can_handle_more_urls_concurrently(self):
        urls = [
            'http://www.google.com',
            'http://www.globo.com',
            'http://www.cnn.com',
            'http://www.bbc.com'
        ]
        otto = Octopus(concurrency=4)

        def handle_url_response(url, response):
            self.responses[url] = response

        for url in urls:
            otto.enqueue(url, handle_url_response)

        otto.start()

        otto.wait(10)

        expect(self.responses).to_length(4)

        for url in urls:
            expect(self.responses).to_include(url)
            expect(self.responses[url].status_code).to_equal(200)

    def test_can_handle_cached_responses(self):
        response = Mock(status_code=200, body="whatever")

        url = 'http://www.google.com'
        otto = Octopus(concurrency=1, cache=True)
        otto.response_cache.put(url, response)

        def handle_url_response(url, response):
            self.response = response

        otto.enqueue(url, handle_url_response)

        expect(self.response).not_to_be_null()
        expect(self.response.status_code).to_equal(200)
        expect(self.response.body).to_equal("whatever")

    def test_can_handle_cached_responses_when_not_cached(self):
        url = 'http://www.google.com'
        otto = Octopus(concurrency=1, cache=True)

        def handle_url_response(url, response):
            self.response = response

        otto.enqueue(url, handle_url_response)
        otto.enqueue(url, handle_url_response)
        otto.enqueue(url, handle_url_response)
        otto.enqueue(url, handle_url_response)

        otto.start()

        otto.wait(2)

        expect(self.response).not_to_be_null()
        expect(self.response.status_code).to_equal(200)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import time

from preggy import expect

from octopus.cache import Cache
from tests import TestCase


class TestCache(TestCase):
    def test_can_create_cache(self):
        cache = Cache(expiration_in_seconds=45)
        expect(cache.expiration_in_seconds).to_equal(45)
        expect(cache.responses).to_be_empty()

    def test_get_returns_none_if_not_put(self):
        cache = Cache(expiration_in_seconds=45)

        expect(cache.get('http://www.google.com')).to_be_null()

    def test_get_returns_none_if_expired(self):
        cache = Cache(expiration_in_seconds=0.1)

        cache.put('http://www.google.com', 'response')

        time.sleep(0.5)

        expect(cache.get('http://www.google.com')).to_be_null()

        expect(cache.responses).not_to_include('http://www.google.com')

    def test_can_get_after_put(self):
        cache = Cache(expiration_in_seconds=10)
        cache.put('http://www.google.com', 'response')
        expect(cache.get('http://www.google.com')).to_equal('response')



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from datetime import datetime, timedelta


class Cache(object):
    def __init__(self, expiration_in_seconds):
        self.responses = {}
        self.expiration_in_seconds = expiration_in_seconds

    def put(self, url, response):
        self.responses[url] = {
            'response': response,
            'expires': datetime.now() + timedelta(seconds=self.expiration_in_seconds)
        }

    def get(self, url):
        if url not in self.responses:
            return None

        data = self.responses[url]

        if data['expires'] <= datetime.now():
            del self.responses[url]
            return None

        return data['response']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from time import time
from threading import Thread

try:
    import requests
except ImportError:
    print("Can't import requests. Probably setup.py installing package.")

try:
    from six.moves import queue
except ImportError:
    print("Can't import six. Probably setup.py installing package.")

from octopus.cache import Cache


class TimeoutError(RuntimeError):
    pass


class OctopusQueue(queue.Queue):
    # from http://stackoverflow.com/questions/1564501/add-timeout-argument-to-pythons-queue-join
    def join_with_timeout(self, timeout):
        self.all_tasks_done.acquire()
        try:
            endtime = time() + timeout
            while self.unfinished_tasks:
                remaining = endtime - time()
                if remaining <= 0.0:
                    raise TimeoutError
                self.all_tasks_done.wait(remaining)
        finally:
            self.all_tasks_done.release()


class Octopus(object):
    def __init__(self, concurrency=10, auto_start=False, cache=False, expiration_in_seconds=30):
        self.concurrency = concurrency
        self.auto_start = auto_start

        self.cache = cache
        self.response_cache = Cache(expiration_in_seconds=expiration_in_seconds)

        self.url_queue = OctopusQueue()

        if auto_start:
            self.start()

    def enqueue(self, url, handler):
        if self.cache:
            response = self.response_cache.get(url)
            if response is not None:
                handler(url, response)
                return

        self.url_queue.put_nowait((url, handler))

    @property
    def queue_size(self):
        return self.url_queue.qsize()

    @property
    def is_empty(self):
        return self.url_queue.empty()

    def start(self):
        for i in range(self.concurrency):
            t = Thread(target=self.do_work)
            t.daemon = True
            t.start()

    def do_work(self):
        while True:
            url, handler = self.url_queue.get()

            response = None
            if self.cache:
                response = self.response_cache.get(url)

            if response is None:
                response = requests.get(url)

                if self.cache:
                    self.response_cache.put(url, response)

            handler(url, response)

            self.url_queue.task_done()

    def wait(self, timeout=10):
        if timeout > 0:
            self.url_queue.join_with_timeout(timeout=timeout)
        else:
            self.url_queue.join()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from octopus import __version__

tests_require = [
    'mock',
    'nose',
    'coverage',
    'yanc',
    'preggy',
    'tox',
    'ipdb',
    'coveralls',
]

setup(
    name='octopus-http',
    version=__version__,
    description='Octopus is a library to use threads to concurrently retrieve and report on the completion of http requests',
    long_description='''
Octopus is a library to use threads to concurrently retrieve and report on the completion of http requests
''',
    keywords='http concurrency threading',
    author='Bernardo Heynemann',
    author_email='heynemann@gmail.com',
    url='https://heynemann.github.io/octopus',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'six',
        'requests'
    ],
    extras_require={
        'tests': tests_require,
    },
)
